/*
** spy_filter_sign.c
** 
** Copyright Julien Perez (C) 2005 
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
** 
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
** 
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <string.h>

#include "spy.h"
#include "http.h"
#include "conf.h"
#include "vshlib.h"

/* 
 * Those structures will organize
 * The signature for fast testing
 *
 * The main Idea is to test easiest first
 */

extern http_session_t http_session;

struct spy_sign_node
{
  void *without;
  void *with_plain;
  void *with_regex;
};
typedef struct spy_sign_node spy_sign_node;


struct spy_sign_leaf
{
  int nb_sign;
  spy_sign_t *head_sign;
  spy_sign_t *tail_sign;
};
typedef struct spy_sign_leaf spy_sign_leaf;


static spy_sign_node content_node;


/* Sign Tree Initilization */

static spy_sign_leaf *create_sign_leaf()
{
  spy_sign_leaf *new_leaf;

  new_leaf = vsh_malloc(sizeof(spy_sign_leaf));
  new_leaf->head_sign = NULL;
  new_leaf->tail_sign = NULL;

  return new_leaf;
}

static void init_url_node(spy_sign_node *node)
{
  node->without     = create_sign_leaf();
  node->with_plain  = create_sign_leaf();
  node->with_regex  = create_sign_leaf();
}

static void init_fld_node(spy_sign_node *node)
{
  node->without     = vsh_malloc(sizeof(spy_sign_node));
  node->with_plain  = vsh_malloc(sizeof(spy_sign_node));
  node->with_regex  = vsh_malloc(sizeof(spy_sign_node));

  init_url_node(node->without);
  init_url_node(node->with_plain);
  init_url_node(node->with_regex);
}

static void init_content_node(spy_sign_node *node)
{
  node->without     = vsh_malloc(sizeof(spy_sign_node));
  node->with_plain  = vsh_malloc(sizeof(spy_sign_node));
  node->with_regex  = vsh_malloc(sizeof(spy_sign_node));

  init_fld_node(node->without);
  init_fld_node(node->with_plain);
  init_fld_node(node->with_regex);
}

int  spy_sign_init ()
{
  init_content_node(&content_node);
  return 1;
}


/* Adding Signature primitive */

static unsigned char hex2uc(char *hex)
{
  unsigned char nbr = 0;

  hex[0] = toupper(hex[0]);
  hex[1] = toupper(hex[1]);

  if (hex[1] <= '9')
    nbr += hex[1] - '0';
  else
    nbr += hex[1] - 'A' + 10;

  if (hex[0] <= '9')
    nbr += (hex[0] - '0') * 16;
  else
    nbr += (hex[0] - 'A' + 10) * 16;
  return nbr;
}

static unsigned char *spy_sign_str(char *str, int *lenght)
{
  int i;
  char hex[2];
  vsh_str_t str_sign;
  unsigned char *sign;

 if (!str || lenght == 0)
    {
      *lenght = 0;
      return NULL;
    }

  if (str[0] != '|')
    {
      *lenght = strlen(str);
      return (unsigned char*) strdup(str);
    }

  vsh_str_init(&str_sign);
  i = 1;
  do {
    
    if (isblank(str[i]))
      i++;
    
    hex[0] = str[i];
    if (str[i + 1] && str[i + 1] != '|')
      hex[1] = str[i + 1];
    else
      {
	*lenght = 0;
	vsh_str_free(&str_sign);
	return NULL;
      }
    vsh_str_addchar(&str_sign, (char)hex2uc(hex));
    i += 2;
  }while (str[i] && str[i] != '|');
  
  if (str[i] == '|')
    {
      *lenght = str_sign.size;
      sign    = vsh_str_inline(&str_sign);
      vsh_str_free(&str_sign);
      return sign;
    }
  *lenght = 0;
  vsh_str_free(&str_sign);
  return NULL;
}


static spy_sign_leaf *spy_sign_findleaf (char *url, char *fld, char *content)     
{
  spy_sign_node *node_fld, *node_url;

  if (content)
    node_fld = content_node.with_plain;
  else
    node_fld = content_node.without;


  if (fld)
    node_url = node_fld->with_plain;
  else
    node_url = node_fld->without;


  if (url)
    return node_url->with_plain;

  return node_url->without;
}

static void spy_sign_addin_leaf(spy_sign_leaf *leaf, spy_sign_t *new_sign)
{
  if (!leaf || !new_sign)
    return ;

  if (leaf->head_sign == NULL)
    {
      leaf->head_sign = new_sign;
      leaf->tail_sign = new_sign;
      return ;
    }
  leaf->tail_sign->next = new_sign;
  leaf->tail_sign = new_sign;
}

static void spy_sign_addsign(char *url, char *fld, 
			     char *msg, char *content, 
			     sign_action action)
{
  spy_sign_leaf *leaf;
  spy_sign_t *new_sign, *tmp_sign;
  

  if (!url && !fld && !msg)
    return ;

  // looking for tha appropriate Leaf
  leaf = spy_sign_findleaf(url, fld, content);

  new_sign = vsh_malloc(sizeof(spy_sign_t));
  new_sign->next = NULL;  
  new_sign->action = action;

  new_sign->url     = NULL;
  new_sign->fld     = NULL;
  new_sign->content = NULL; 
  new_sign->msg     = NULL; 

  if (url)
    new_sign->url = spy_sign_str(url, &(new_sign->size_url));
  if (fld)
    new_sign->fld = spy_sign_str(fld, &(new_sign->size_fld));
  if (content)
    new_sign->content = spy_sign_str(content, &(new_sign->size_content));
  if (msg)
    new_sign->msg = strdup(msg);
 


  // adding the signature in the Tree
  spy_sign_addin_leaf(leaf, new_sign);
}



int  spy_sign_policy_rule (char *rule)
{
  int i, ok;
  vsh_str_t str;
  char *content;

  char *url, *fld, *msg, *cnt;

  i = 0;
  url  = NULL;
  fld  = NULL;
  msg  = NULL;
  cnt  = NULL;

  while(i < strlen(rule))
    {
     /* trim the rule*/
      for (i; rule[i] && isblank(rule[i]); i++);
      if (!rule[i])
	continue ;

      /* grab the field : "url" | "fld" | "msg" */
      vsh_str_init(&str);
      while (rule[i] && !isblank(rule[i]) && rule[i] != '"')
	{
	  vsh_str_addchar(&str, rule[i]);
	  i++;
	}
      content = vsh_str_inline(&str);
      vsh_str_free(&str);

      /* grab the value of the field */
      while (rule[i] && rule[i] != '"')
	i++ ;

      if (rule[i] && rule[i] == '"')
	i++;
      else
	{
	  perror("spy_sign_policy_rule");
	  exit(1);
	}

      vsh_str_init(&str);
      while (rule[i] && rule[i] != '"')
	{
	  vsh_str_addchar(&str, rule[i]);
	  i++;
	}

      ok = 0;
      if (!strcmp("url", content))
	{
	  url = vsh_str_inline(&str);
	  ok = 1;
	}
      if (!strcmp("fld", content))
	{
	  fld = vsh_str_inline(&str);
	  ok = 1;
	}

      if (!strcmp("content", content))
	{
	  cnt = vsh_str_inline(&str);
	  ok = 1;
	}

      if (!strcmp("msg", content))
	{
	  msg = vsh_str_inline(&str);
	  ok = 1;
	}

      vsh_free(content);
      vsh_str_free(&str);

      if (rule[i])
	i++;

      if (!ok)
	break ;
    }

  /* Insert the new signature */
  spy_sign_addsign(url, fld, msg, cnt, SIGN_ACT_BLOCK);
  vsh_free(url);
  vsh_free(fld);
  vsh_free(msg);
  vsh_free(cnt);
  return 1;
}




/* Signature Based Checking algorithm */

static int spy_sign_strstr(unsigned char *str, unsigned char *needle, int size_needle, int size_str)
{
  int i, j, k;

  //printf("spy_sign_strstr [%s] [%s] [%d]\n", str, needle, size_needle);

  if (strlen(str) < size_needle)
    return -1;

  if (!size_needle)
    return -1;

  for (i = 0; i < size_str; i++)
    {
      for (k = i, j = 0; j < size_needle && k < size_str && needle[j] == str[k]; k++, j++) ;
      //printf("j [%d]\n", j);
      if (j == size_needle)
	return i;
    }
  return -1;
}

static spy_sign_t *spy_sign_match_lst (spy_sign_t *lst_sign, char *url, http_field_t *lst_fields, unsigned char *content, int content_lenght)
{
  int res;
  spy_sign_t    *tmp_sign;
  http_field_t  *tmp_field;

  for (tmp_sign = lst_sign; tmp_sign; tmp_sign = tmp_sign->next)
    { 
      res = 0;

      if (tmp_sign->url != NULL)
	if (spy_sign_strstr(url, tmp_sign->url, tmp_sign->size_url, strlen(url)) == -1)
	  continue ;
      
      if (tmp_sign->fld)
	{
	  tmp_field = lst_fields;
	  do {
	    res = spy_sign_strstr(tmp_field->value, tmp_sign->fld, tmp_sign->size_fld, strlen(tmp_field->value));
	    if (res == -1)
	      tmp_field = tmp_field->next;
	  }while (!res && tmp_field);
	}

      if (res == -1)
	continue ;

      if (tmp_sign->size_content)
	{	
	  if (spy_sign_strstr(content, tmp_sign->content, tmp_sign->size_content, content_lenght) != -1)
	    return tmp_sign ;
	}
      else
	return tmp_sign ;
    }
  return NULL;
}

static spy_sign_t *spy_sign_match_url (spy_sign_node *node, char *url, http_field_t *lst_fields, unsigned char *content, int content_lenght)
{
  spy_sign_leaf *leaf;
  spy_sign_t *sign;
  
  // printf("spy_sign_match_url\n");
  if (url)
    {
      leaf = node->with_plain;
      sign = spy_sign_match_lst (leaf->head_sign, url, lst_fields, content, content_lenght);
      if (sign)
	return sign;
    }
  leaf = node->without;
  return spy_sign_match_lst  (leaf->head_sign, url, lst_fields, content, content_lenght);
}

static spy_sign_t *spy_sign_match_fld(spy_sign_node *node, char *url, http_field_t *lst_fields, unsigned char *content, int content_lenght)
{
  spy_sign_t *sign;

  // printf("spy_sign_match_fld\n");
  if (lst_fields)
     {
       sign = spy_sign_match_url (node->with_plain, url, lst_fields, content, content_lenght);
       if (sign)
	 return sign;
     }
   return spy_sign_match_url  (node->without, url, lst_fields, content, content_lenght);
}

static spy_sign_t *spy_sign_match(char *url, http_field_t *lst_fields, unsigned char *content, int content_lenght)
{
  spy_sign_t *sign;
  // printf("spy_sign_match\n");
  
  /* select content node*/
  if (content_lenght)
    {
      sign = spy_sign_match_fld (content_node.with_plain, url, lst_fields, content, content_lenght);
      if (sign)
	return sign;
    }
  return spy_sign_match_fld (content_node.without, url, lst_fields, content, content_lenght);
}



/* 
 * Public Interface  
 */

/* Headers Filtering */

int  spy_sign_req  (http_req_t *req, int log)
{
  spy_sign_t *sign;
  char       *log_msg;

  // printf("spy_sign_req\n");
  if ((sign = spy_sign_match(req->url, req->lst_fields, NULL, 0)) != NULL)
    {
      if (log)
	{
	  log_msg = vsh_malloc(strlen(conf_get_node(CONF_LOG)) + strlen(sign->msg) + 2);
	  sprintf(log_msg, "%s %s", conf_get_node(CONF_LOG), sign->msg);
	  spy_log("BLOCK_SIGN", log_msg, 1);
	  vsh_free(log_msg);
	}
      return 0;
    }
  // printf("end spy_sign_req\n");
  return 1;
}


int  spy_sign_resp (http_resp_t *resp, int log)
{
  spy_sign_t *sign;
  char       *log_msg;

  if ((sign = spy_sign_match(NULL, resp->lst_fields, NULL, 0)) != NULL)
    {
      if (log)
	{
	  log_msg = vsh_malloc(strlen(conf_get_node(CONF_LOG)) + strlen(sign->msg) + 2);
	  sprintf(log_msg, "%s %s", conf_get_node(CONF_LOG), sign->msg);
	  spy_log("BLOCK_SIGN", log_msg, 1);
	  vsh_free(log_msg);
	}
      return 0;
    }
  return 1;
}


/* Content filtering */

static char *tmp_content_type;
static int tmp_flow ;
int spy_sign_begin_session(char *content_type, int flow)
{
  tmp_content_type = content_type;
  tmp_flow = flow;
}


int spy_sign_content (void *data, int lenght, int log)
{
  spy_sign_t *sign;
  char       *log_msg;

  if (lenght <= 0)
    return 1;

  

  if (tmp_flow == POLICY_INFLOW)
    {
      if ((sign = spy_sign_match(NULL, http_session.resp_srv.lst_fields, data, lenght)) != NULL)
	{
	  if (log)
	    {
	      log_msg = vsh_malloc(strlen(conf_get_node(CONF_LOG)) + strlen(sign->msg) + 2);
	      sprintf(log_msg, "%s %s", conf_get_node(CONF_LOG), sign->msg);
	      spy_log("BLOCK_SIGN", log_msg, 1);
	      vsh_free(log_msg);
	    }
	  return 0;
	}
    }
  else
    {
      if ((sign = spy_sign_match(http_session.req_cli.url, http_session.req_cli.lst_fields, data, lenght)) != NULL)
	{
	  if (log)
	    {
	      log_msg = vsh_malloc(strlen(conf_get_node(CONF_LOG)) + strlen(sign->msg) + 2);
	      sprintf(log_msg, "%s %s", conf_get_node(CONF_LOG), sign->msg);
	      spy_log("BLOCK_SIGN", log_msg, 1);
	      vsh_free(log_msg);
	    }
	  return 0;
	}
    }
  return 1;
}
